@@ -41,7 +41,7 @@ class CallbackArgument:
 class Method:
     def __init__(self, cls, name, c_name,
             return_semantics, return_base_type, return_ptr,
-            args):
+            args, is_exported, is_constructor):
         self.cls = cls
         self.name = name
         self.c_name = c_name
@@ -50,6 +50,8 @@ class Method:
         self.return_ptr = return_ptr
         self.args = args
         self.mutator_veto = False
+        self.is_exported = is_exported
+        self.is_constructor = is_constructor
 
         if not self.is_static:
             self.args[0].name = "self"
@@ -74,18 +76,24 @@ class Method:
 
 
 
-CLASSES = [
-        "basic_set_list", "set_list", "aff_list", "band_list",
-        "printer",  "mat", "vec",
+PART_TO_CLASSES = {
+        "part1": [
+        "basic_set_list", "set_list", "aff_list", "pw_aff_list", "band_list",
+        "printer",  "mat", "vec", "id",
         "aff", "pw_aff",
+        "multi_aff", "pw_multi_aff",
 
-         "div", "dim",  "constraint", "local_space",
+         "constraint", "space", "local_space",
+         ],
 
+        "part2": [
         "basic_set", "basic_map",
         "set", "map",
         "union_map", "union_set",
-        "point", "vertex", "cell", "vertices", "dim_set",
+        "point", "vertex", "cell", "vertices",
+        ],
 
+        "part3": [
         "qpolynomial_fold", "pw_qpolynomial_fold",
         "union_pw_qpolynomial_fold",
         "union_pw_qpolynomial", "term",
@@ -95,6 +103,10 @@ CLASSES = [
 
         "access_info", "flow",
         ]
+        }
+CLASSES = []
+for cls_list in PART_TO_CLASSES.itervalues():
+    CLASSES.extend(cls_list)
 
 CLASS_MAP = {
         "equality": "constraint",
@@ -122,7 +134,7 @@ FUNC_PTR_RE = re.compile(r"""
     \)
     """,
     re.VERBOSE)
-STRUCT_DECL_RE = re.compile(r"struct\s+([a-z_A-Z0-9]+)\s*;")
+STRUCT_DECL_RE = re.compile(r"(__isl_export\s+)?struct\s+([a-z_A-Z0-9]+)\s*;")
 ARG_RE = re.compile(r"^((?:\w+)\s+)+(\**)\s*(\w+)$")
 INLINE_SEMICOLON_RE = re.compile(r"\;[ \t]*(?=\w)")
 
@@ -310,16 +322,23 @@ class FunctionData:
 
             else:
                 decl = ""
+
                 while True:
                     decl = decl + l
+                    if decl:
+                        decl += " "
                     i += 1
+                    if STRUCT_DECL_RE.search(decl):
+                        break
+
                     open_par_count = sum(1 for i in decl if i == "(")
                     close_par_count = sum(1 for i in decl if i == ")")
-                    if open_par_count == close_par_count:
+                    if open_par_count and open_par_count == close_par_count:
                         break
                     l = lines[i].strip()
 
-                self.parse_decl(decl)
+                if not STRUCT_DECL_RE.search(decl):
+                    self.parse_decl(decl)
 
     def parse_decl(self, decl):
         decl_match = DECL_RE.match(decl)
@@ -364,6 +383,15 @@ class FunctionData:
             name = name + "_"
 
         words = return_base_type.split()
+
+        is_exported = "__isl_export" in words
+        if is_exported:
+            words.remove("__isl_export")
+
+        is_constructor = "__isl_constructor" in words
+        if is_constructor:
+            words.remove("__isl_constructor")
+
         return_semantics, words = filter_semantics(words)
         words = [w for w in words if w not in ["struct", "enum"]]
         return_base_type = " ".join(words)
@@ -372,7 +400,7 @@ class FunctionData:
         cls_meth_list.append(Method(
                 cls, name, c_name,
                 return_semantics, return_base_type, return_ptr,
-                args))
+                args, is_exported=is_exported, is_constructor=is_constructor))
 
 
 
@@ -551,6 +579,8 @@ def write_wrapper(outf, meth):
             docs.append(":param %s: integer" % arg.name)
 
         elif arg.base_type.startswith("isl_") and arg.ptr == "*":
+            need_nonconst = False
+
             arg_cls = arg.base_type[4:]
             arg_descr = ":param %s: :class:`%s`" % (arg.name, to_py_class(arg_cls))
 
@@ -558,7 +588,7 @@ def write_wrapper(outf, meth):
                 raise Undocumented(meth)
 
             checks.append("""
-                if (!arg_%(name)s || !arg_%(name)s->is_valid())
+                if (!arg_%(name)s.is_valid())
                   PYTHON_ERROR(ValueError, "passed invalid arg to isl_%(meth)s for %(name)s");
                 """ % dict(name=arg.name, meth="%s_%s" % (meth.cls, meth.name)))
 
@@ -566,11 +596,11 @@ def write_wrapper(outf, meth):
             if arg.semantics is SEM_TAKE:
                 if copyable:
                     checks.append("""
-                        if (!arg_%(name)s || !arg_%(name)s->is_valid())
+                        if (!arg_%(name)s.is_valid())
                           PYTHON_ERROR(ValueError, "passed invalid arg to isl_%(meth)s for %(name)s");
                         std::auto_ptr<%(cls)s> auto_arg_%(name)s;
                         {
-                            isl_%(cls)s *tmp_ptr = isl_%(cls)s_copy(arg_%(name)s->m_data);
+                            isl_%(cls)s *tmp_ptr = isl_%(cls)s_copy(arg_%(name)s.m_data);
                             if (!tmp_ptr)
                                 throw std::runtime_error("failed to copy arg %(name)s on entry to %(meth)s");
                             auto_arg_%(name)s = std::auto_ptr<%(cls)s>(new %(cls)s(tmp_ptr));
@@ -584,19 +614,24 @@ def write_wrapper(outf, meth):
                     passed_args.append("auto_arg_%s->m_data" % arg.name)
 
                 else:
-                    if not (arg_idx == 0 and meth.is_mutator):
-                        post_call.append("arg_%s->invalidate();" % arg.name)
+                    need_nonconst = True
 
-                    passed_args.append("arg_%s->m_data" % arg.name)
+                    if not (arg_idx == 0 and meth.is_mutator):
+                        post_call.append("arg_%s.invalidate();" % arg.name)
+
+                    passed_args.append("arg_%s.m_data" % arg.name)
 
                     if arg_idx == 0 and meth.is_mutator:
                         arg_descr += " (mutated in-place)"
                     else:
                         arg_descr += " (:ref:`becomes invalid <auto-invalidation>`)"
             else:
-                passed_args.append("arg_%s->m_data" % arg.name)
+                passed_args.append("arg_%s.m_data" % arg.name)
 
-            input_args.append("%s *%s" % (arg_cls, "arg_"+arg.name))
+            if need_nonconst:
+                input_args.append("%s &%s" % (arg_cls, "arg_"+arg.name))
+            else:
+                input_args.append("%s const &%s" % (arg_cls, "arg_"+arg.name))
 
             docs.append(arg_descr)
 
@@ -660,9 +695,14 @@ def write_wrapper(outf, meth):
         ret_descr = processed_return_type
 
         if extra_ret_vals:
-            processed_return_type = "py::object"
-            body.append("return py::make_tuple(result, %s);" % ", ".join(extra_ret_vals))
-            ret_descr = "tuple: (%s, %s)" % (ret_descr, ", ".join(extra_ret_descrs))
+            if len(extra_ret_vals) == 1:
+                processed_return_type = "py::object"
+                body.append("return py::object(%s);" % extra_ret_vals[0])
+                ret_descr = extra_ret_descrs[0]
+            else:
+                processed_return_type = "py::object"
+                body.append("return py::make_tuple(%s);" % ", ".join(extra_ret_vals))
+                ret_descr = "tuple: (%s)" % (", ".join(extra_ret_descrs))
         else:
             body.append("return result;")
 
@@ -683,8 +723,8 @@ def write_wrapper(outf, meth):
                 meth.mutator_veto = True
                 raise Retry()
 
-            processed_return_type = "isl::%s *" % ret_cls
-            body.append("arg_%s->m_data = result;" % meth.args[0].name)
+            processed_return_type = "isl::%s &" % ret_cls
+            body.append("arg_%s.m_data = result;" % meth.args[0].name)
             body.append("return arg_%s;" % meth.args[0].name)
 
             ret_descr = ":class:`%s` (self)" % to_py_class(ret_cls)
@@ -797,9 +837,6 @@ def write_exposer(outf, meth, arg_names, doc_str, static_decls):
 
     extra_py_names = []
 
-    if meth.name == "is_equal":
-        extra_py_names.append("__eq__")
-
     #if meth.is_static:
         #doc_str = "(static method)\n" + doc_str
 
@@ -826,6 +863,12 @@ def write_wrappers(expf, wrapf, methods):
     static_decls = []
 
     for meth in methods:
+        if meth.name.endswith("_si") or meth.name.endswith("_ui") and len([
+            meth2.name == meth.name[:-3]
+            for meth2 in methods]):
+            # no need to expose C integer versions of things
+            continue
+
         try:
             arg_names, doc_str = write_wrapper(wrapf, meth)
             write_exposer(expf, meth, arg_names, doc_str, static_decls)
@@ -850,7 +893,7 @@ def write_wrappers(expf, wrapf, methods):
 def gen_wrapper(include_dirs):
     fdata = FunctionData(["."] + include_dirs)
     fdata.read_header("isl_list.h")
-    fdata.read_header("isl/dim.h")
+    fdata.read_header("isl/space.h")
     fdata.read_header("isl/set.h")
     fdata.read_header("isl/map.h")
     fdata.read_header("isl/local_space.h")
@@ -867,18 +910,19 @@ def gen_wrapper(include_dirs):
     fdata.read_header("isl/band.h")
     fdata.read_header("isl/schedule.h")
     fdata.read_header("isl/flow.h")
-    fdata.read_header("isl/div.h")
 
-    expf = open("src/wrapper/gen-expose.inc", "wt")
-    wrapf = open("src/wrapper/gen-wrap.inc", "wt")
 
-    write_wrappers(expf, wrapf, [
-        meth
-        for methods in fdata.classes_to_methods.itervalues()
-        for meth in methods])
+    for part, classes in PART_TO_CLASSES.iteritems():
+        expf = open("src/wrapper/gen-expose-%s.inc" % part, "wt")
+        wrapf = open("src/wrapper/gen-wrap-%s.inc" % part, "wt")
 
-    expf.close()
-    wrapf.close()
+        write_wrappers(expf, wrapf, [
+            meth
+            for cls in classes
+            for meth in fdata.classes_to_methods.get(cls, [])])
+
+        expf.close()
+        wrapf.close()
 
 
 
@@ -886,4 +930,4 @@ def gen_wrapper(include_dirs):
 
 if __name__ == "__main__":
     from os.path import expanduser
-    gen_wrapper([expanduser("~/pool/include")])
+    gen_wrapper([expanduser("isl/include")])
